/*
The software in this package is distributed under the GNU General
Public License version 2 (with a special exception described below).

A copy of GNU General Public License (GPL) is included in this distribution,
in the file COPYING.GPL.

As a special exception, if other files instantiate templates or use macros
or inline functions from this file, or you compile this file and link it
with other works to produce a work based on this file, this file
does not by itself cause the resulting work to be covered
by the GNU General Public License.

However, the source code for this file must still be made available
in accordance with section (3) of the GNU General Public License.

This exception does not invalidate any other reasons why a work based
on this file might be covered by the GNU General Public License.
*/
/** @file
 * @brief netstring related functions implementations.
 *
 * @copyright Copyright &copy; 2024 by Intra2net AG
 *
 */

#include <netstring.hpp>
#include <stringfunc.hxx>
#include <cstdio>
#include "exception.hxx"
#include "boost/lexical_cast.hpp"

using namespace I2n;

/**
 * @brief Constructor for NetstringBuffer
 */
NetstringBuffer::NetstringBuffer() :
        SizeLimit (1073741824), //1gb
        ExpectedBytes(0),
        NumBytesRead (0),
        LengthBeforeColon (0),
        BuffCursor(0),
        CurrentState(ReadLength)
        {}

/**
 * @brief Constructor for NetstringBuffer
 * @param size_limit Maximum package size (recommended = 1073741824)
 */
NetstringBuffer::NetstringBuffer(size_t size_limit) :
        SizeLimit (size_limit),
        ExpectedBytes(0),
        NumBytesRead (0),
        LengthBeforeColon (0),
        BuffCursor(0),
        CurrentState(ReadLength)
        {}

/**
 * @brief Extract a full and validated netstring from the buffer
 *
 * If this function throws an exception, the object may be in an unclean state,
 * so should not be used any more.
 *
 * @note Assumes that there is a validated netstring on the buffer
 *       Should only be called by the append function
 * @return string decoded from the buffer
 * @throws netstring_error if the buffer is empty or contains invalid data
 */
std::string NetstringBuffer::get_next_from_buffer()
{
    if(!this->InputBuffer.empty())
    {
        return decode_string(this->InputBuffer);
    }
    else
        throw netstring_error("ERROR: trying to decode string from empty buffer");
}

/**
 * @brief Try to read the length of a netstring
 *
 * If this function throws an exception, the object may be in an unclean state,
 * so should not be used any more.
 *
 * @note Should be only called by append_to_buf()
 * @throws netstring_error - input had an invalid format; do not use object afterwards.
 */
void NetstringBuffer::parse_read_length()
{
    // we can assume here, that the buffer starts with a new encoded netstring,
    // there can be no left-overs from previous netstrings at the start of the buffer
    string::size_type colon_i = InputBuffer.find_first_of(':');

    if (colon_i != std::string::npos)
    {
        // found colon, read length and start reading bytes
        try
        {
            string length_part = this->InputBuffer.substr(0, colon_i);
            this->ExpectedBytes = boost::lexical_cast<size_t>(length_part);

            this->CurrentState = ReadNBytes;
            this->LengthBeforeColon = colon_i + 1;
            this->BuffCursor = this->LengthBeforeColon;

            if (this->ExpectedBytes > this->SizeLimit)
            {
                throw netstring_error("ERROR: data package to big");
            }
        }
        // didn't find the number of bytes (invalid)
        catch (boost::bad_lexical_cast &)
        {
            throw netstring_error("ERROR: invalid netstring, format for length should be '%d:%s,'");
        }

    }

    // no colon: verify if there's only digits
    else
    {
        try
        {
            size_t num = boost::lexical_cast<size_t>(this->InputBuffer);
            if (num > this->SizeLimit)
            {
                throw netstring_error("ERROR: data package to big");
            }
        }
        catch (boost::bad_lexical_cast &)
        {
            throw netstring_error("ERROR: invalid netstring, should only have digits before colon");
        }
        this->BuffCursor = InputBuffer.length();
        //continue waiting for colon and receiving only digits
    }
}

/**
 * @brief Try to read the bytes part of the netstring
 *
 * @note Should be only called by append_to_buf()
 */
void NetstringBuffer::parse_read_n_bytes()
{
    size_t str_len = (this->InputBuffer.length() - this->BuffCursor);

    // has more than one netstring
    if (str_len > (this->ExpectedBytes - this->NumBytesRead + 1 ) ) {
        this->NumBytesRead = this->NumBytesRead + this->ExpectedBytes;
        this->add_cursor(this->ExpectedBytes);
    }
    // only bytes of one netstring
    else
    {
        this->NumBytesRead = this->NumBytesRead + str_len;
        this->add_cursor(str_len);
    }

    // finished reading everything
    if (this->NumBytesRead >= this->ExpectedBytes)
    {
        this->CurrentState = ReadTerminatingChar;
    }
}

/**
 * @brief Try to read the terminating char (comma) ','
 *
 * If this function throws an exception, the object may be in an unclean state,
 * so should not be used any more.
 *
 * Removes one complete encoded netstring from the buffer, decodes it and pushes it
 * onto the ValidNetstrings.
 *
 * @note Should be only called by append_to_buf()
 * @throws netstring_error - input had an invalid format; do not use object afterwards.
 */
void NetstringBuffer:: parse_read_terminating_char()
{
    if (this->InputBuffer.find_last_of(',',this->BuffCursor) != std::string::npos) {
        if (this->InputBuffer[(this->LengthBeforeColon + this->ExpectedBytes)] == ',') {
            string valid_str = this->get_next_from_buffer();
            this->ValidNetstrings.push_back(valid_str);

            this->InputBuffer = this->InputBuffer
                    .replace(0, (this->LengthBeforeColon + this->ExpectedBytes + 1), "");

            this->add_cursor(1); // ','
            this->reset_state_machine();
        }
        else
        {
            throw netstring_error("ERROR: invalid netstring, should always end with comma");
        }
    }
    // If already read more then the size (should have read comma)
    // and still missing a comma
    else if( this->InputBuffer.length() > this->BuffCursor
            || this->NumBytesRead > this->ExpectedBytes)
    {
        throw netstring_error("ERROR: invalid netstring, should always end with comma");
    }
}

/**
 * @brief Append a string(bytes) to buffer and check if there is a
 *        valid netstring already, if so will extract it and put it
 *        on to the deque for valid data.
 *
 * If this function throws an exception, the object may be in an unclean state,
 * so should not be used any more.
 *
 * @param str - new bytes to append
 * @throws netstring_error - input had an invalid format; do not use object afterwards.
 */
void NetstringBuffer::append_to_buf(const string& str)
{
    if(str.empty())
    {
        return;
    }
    this->InputBuffer.append(str);
    do {
        if (this->CurrentState == ReadLength) {
            parse_read_length();
        }
        if (this->CurrentState == ReadNBytes) {
            parse_read_n_bytes();
        }
        if (this->CurrentState == ReadTerminatingChar) {
            parse_read_terminating_char();
        }
    } while (this->InputBuffer.length() > this->BuffCursor);

}

/**
 * @brief Encode string in the netstring format:
 *        '<len>:<bytes>,'
 *
 * @param original - string to be encoded
 * @return string encoded in netstring format
 */
string NetstringBuffer::encode_string(const string& original)
{
    unsigned int len = original.length();
    return std::string((to_string(len)) + ":" + original + ",");
}

/**
 * @brief Decodes a string encoded in the netstring format
 *
 * @note Assuming ALWAYS that the string is valid
 * @param original - string with encoded in netstring
 * @return string decoded
 * @throws netstring_error if the input is not valid netstring format
 */
string NetstringBuffer::decode_string(const string &original)
{
    int len, colon_i = original.find_first_of(":");

    if(sscanf(original.c_str(),"%d:", &len) != 1 || colon_i == std::string::npos)
        throw netstring_error("ERROR: invalid netstring format, impossible to decode");

    return original.substr(colon_i + 1,len);
}

/**
 * @brief Reset state machine to read another netstring from buffer
 */
void NetstringBuffer::reset_state_machine()
{
    this->CurrentState = ReadLength;
    this->NumBytesRead = 0;
    this->ExpectedBytes = -1;
    this->LengthBeforeColon = 0;
    this->BuffCursor = 0;
}

/**
 * @brief Returns validated data from the netstring buffer
 *
 * @return string - data extracted from netstring buffer
 *         or empty string if there are no more strings
 *         in the buffer
 */
string NetstringBuffer::pop_data()
{
    if(ValidNetstrings.empty())
        return "";

    string str = ValidNetstrings.front();
    ValidNetstrings.pop_front();
    return str;
}