/*
The software in this package is distributed under the GNU General
Public License version 2 (with a special exception described below).

A copy of GNU General Public License (GPL) is included in this distribution,
in the file COPYING.GPL.

As a special exception, if other files instantiate templates or use macros
or inline functions from this file, or you compile this file and link it
with other works to produce a work based on this file, this file
does not by itself cause the resulting work to be covered
by the GNU General Public License.

However the source code for this file must still be made available
in accordance with section (3) of the GNU General Public License.

This exception does not invalidate any other reasons why a work based
on this file might be covered by the GNU General Public License.
*/
/** @file
 * @brief netstring related functions and classes.
 *
 * @copyright Intra2net AG 2024
 */

#ifndef LIBI2NCOMMON_NETSTRING_H
#define LIBI2NCOMMON_NETSTRING_H

#include <string>
#include <deque>

using namespace std;

namespace I2n {

/**
 * @brief Small state machine:
 *        .-> 1 - ReadLength -v
 *        '   2 - ReadNBytes <'   --v
 *        '-- 3 - ReadTerminatingChar ','  -> move to ValidNetstrings, remove from InputBuffer via
 *                                            .append(), switch to ReadLength state.
 */
enum NetstringStates
{
    ReadLength = 0,
    ReadNBytes = 1,
    ReadTerminatingChar = 2,
};

/**
 * @brief Class to control netstring buffering and extraction
 *
 */
class NetstringBuffer {

    private:
        size_t SizeLimit;
        size_t ExpectedBytes;
        size_t NumBytesRead;
        size_t LengthBeforeColon;
        size_t BuffCursor;
        NetstringStates CurrentState;

        deque<string> ValidNetstrings;
        string InputBuffer;

        string get_next_from_buffer();

        void add_cursor(const size_t size)
        {
            BuffCursor+=size;
        }

        static string decode_string(const string &str);

        void parse_read_length();

        void parse_read_n_bytes();

        void parse_read_terminating_char();

    public:
        NetstringBuffer();

        NetstringBuffer(size_t size_limit);


        size_t get_amount_of_valid_strings() const
        {
            return ValidNetstrings.size();
        }

        string pop_data();

        void append_to_buf(const string& str);

        static string encode_string(const string &original);

        void reset_state_machine();

};

} // namespace I2n

#endif //LIBI2NCOMMON_NETSTRING_H