/*
The software in this package is distributed under the GNU General
Public License version 2 (with a special exception described below).

A copy of GNU General Public License (GPL) is included in this distribution,
in the file COPYING.GPL.

As a special exception, if other files instantiate templates or use macros
or inline functions from this file, or you compile this file and link it
with other works to produce a work based on this file, this file
does not by itself cause the resulting work to be covered
by the GNU General Public License.

However, the source code for this file must still be made available
in accordance with section (3) of the GNU General Public License.

This exception does not invalidate any other reasons why a work based
on this file might be covered by the GNU General Public License.
*/
/** @file
 * @brief unit test for the netstring functionalities.
 *
 * @copyright &copy; Copyright 2024 Intra2net AG
 *
 */
#include <boost/test/unit_test_suite.hpp>
#include <boost/test/test_tools.hpp>
#include <string>

#include "netstring.hpp"
#include "stringfunc.hxx"
#include "exception.hxx"

using namespace I2n;

BOOST_AUTO_TEST_SUITE(TestNetstring)

BOOST_AUTO_TEST_CASE(TestEnconding)
{
    string test1 = "hallo?_this!^#&$(-=+*&^%$#@!/>?<_is a test:{><>?P}";

    BOOST_CHECK_EQUAL("50:hallo?_this!^#&$(-=+*&^%$#@!/>?<_is a test:{><>?P},", NetstringBuffer::encode_string(test1));
}

BOOST_AUTO_TEST_CASE(TestEncondingEmpty)
{
    string test1 = "";

    BOOST_CHECK_EQUAL(I2n::to_string(test1.length())+":"+test1+",", NetstringBuffer::encode_string(test1));
    BOOST_CHECK_EQUAL(0, atoi(&(NetstringBuffer::encode_string(test1)[0])));
    BOOST_CHECK_EQUAL(3,NetstringBuffer::encode_string(test1).length());
    BOOST_CHECK_EQUAL("0:,",NetstringBuffer::encode_string(test1));
}

BOOST_AUTO_TEST_CASE(TestAppendingToBuffer)
{
    NetstringBuffer netbuff = NetstringBuffer();

    netbuff.append_to_buf("10:testtest12,");

    BOOST_CHECK_EQUAL(1,netbuff.get_amount_of_valid_strings());
}

BOOST_AUTO_TEST_CASE(TestAutoDecoding)
{
    string test = "testtest12";

    NetstringBuffer netbuff = NetstringBuffer();

    netbuff.append_to_buf(NetstringBuffer::encode_string(test));

    BOOST_CHECK_EQUAL("testtest12", netbuff.pop_data());
}

BOOST_AUTO_TEST_CASE(TestAutoDecodingMoreThenOne)
{
    string test = "testtest12";
    string test2 = "123teste123";

    NetstringBuffer netbuff = NetstringBuffer();

    netbuff.append_to_buf(NetstringBuffer::encode_string(test));
    netbuff.append_to_buf(NetstringBuffer::encode_string(test2));

    BOOST_CHECK_EQUAL(test, netbuff.pop_data());
    BOOST_CHECK_EQUAL(test2, netbuff.pop_data());
}

BOOST_AUTO_TEST_CASE(TestAutoDecodingThree)
{
    NetstringBuffer netbuff = NetstringBuffer();

    string test = "testtest12";
    string test2 = "123teste123";
    string test3 = "hallo";

    netbuff.append_to_buf(NetstringBuffer::encode_string(test));
    netbuff.append_to_buf(NetstringBuffer::encode_string(test2));
    netbuff.append_to_buf(NetstringBuffer::encode_string(test3));

    BOOST_CHECK_EQUAL(test, netbuff.pop_data());
    BOOST_CHECK_EQUAL(test2, netbuff.pop_data());
    BOOST_CHECK_EQUAL(test3, netbuff.pop_data());
}

BOOST_AUTO_TEST_CASE(TestSingleAppend)
{
    NetstringBuffer netbuff = NetstringBuffer();

    netbuff.append_to_buf("10:testtest12,10:othertest1,");

    BOOST_CHECK_EQUAL("testtest12", netbuff.pop_data());
    BOOST_CHECK_EQUAL("othertest1", netbuff.pop_data());
}

BOOST_AUTO_TEST_CASE(TestHalfNetstring)
{
    NetstringBuffer netbuff = NetstringBuffer();

    netbuff.append_to_buf("10:test");
    netbuff.append_to_buf("test12,");

    BOOST_CHECK_EQUAL("testtest12", netbuff.pop_data());
}

BOOST_AUTO_TEST_CASE(TestFormat_Comma)
{
    NetstringBuffer netbuff = NetstringBuffer();

    netbuff.append_to_buf("10:testtest12");

    string test2 = "10:othertest1,";
    BOOST_CHECK_THROW(netbuff.append_to_buf(test2), netstring_error)
}

BOOST_AUTO_TEST_CASE(TestFormat_CommaBreak)
{
    NetstringBuffer netbuff = NetstringBuffer();

    netbuff.append_to_buf("10:test");
    netbuff.append_to_buf("test12");

    string test2 = "10:othertest1,";
    BOOST_CHECK_THROW(netbuff.append_to_buf(test2), netstring_error)
}

BOOST_AUTO_TEST_CASE(TestFormat_Colon)
{
    string test = "10test";
    NetstringBuffer netbuff = NetstringBuffer();
    try
    {
        netbuff.append_to_buf(test);
    }
    catch (netstring_error &e )
    {
        BOOST_CHECK_EQUAL(e.what(),"ERROR: invalid netstring, should only have digits before colon");
    }
}

BOOST_AUTO_TEST_CASE(TestFormat_NoNumber)
{
    string test = ":test";
    NetstringBuffer netbuff = NetstringBuffer();
    try
    {
        netbuff.append_to_buf(test);
    }
    catch (netstring_error &e )
    {
        BOOST_CHECK_EQUAL(e.what(),"ERROR: invalid netstring, format for length should be '%d:%s,'");
    }
}

BOOST_AUTO_TEST_CASE(TestSendByteByByte)
{
    NetstringBuffer netbuff = NetstringBuffer();

    netbuff.append_to_buf("5");
    netbuff.append_to_buf(":");
    netbuff.append_to_buf("H");
    netbuff.append_to_buf("a");
    netbuff.append_to_buf("l");
    netbuff.append_to_buf("l");
    netbuff.append_to_buf("o");
    netbuff.append_to_buf(",");

    BOOST_CHECK_EQUAL("Hallo", netbuff.pop_data());
}

BOOST_AUTO_TEST_CASE(TestSendByteByByteTwo)
{
    NetstringBuffer netbuff = NetstringBuffer();
    string test = "5:Hallo,3:Ola,";

    for (size_t i = 0; i < test.length(); ++i)
    {
        netbuff.append_to_buf(I2n::to_string(test[i]));
    }

    BOOST_CHECK_EQUAL("Hallo", netbuff.pop_data());
    BOOST_CHECK_EQUAL("Ola", netbuff.pop_data());
}

BOOST_AUTO_TEST_CASE(TestPkgOverflow)
{
    NetstringBuffer netbuff = NetstringBuffer();
    string test = "1073741826:"; //  bigger then 1073741824
    BOOST_CHECK_THROW(netbuff.append_to_buf(test), netstring_error)
}

BOOST_AUTO_TEST_CASE(TestAppendingEmptyBuffer)
{
    NetstringBuffer netbuff = NetstringBuffer();
    netbuff.append_to_buf("");
    BOOST_CHECK(true);
}

BOOST_AUTO_TEST_CASE(TestAppendingEmptyBufferAndNetstrings)
{
    NetstringBuffer netbuff = NetstringBuffer();
    netbuff.append_to_buf("");
    netbuff.append_to_buf("5:Hallo,3:Ola,");
    BOOST_CHECK_EQUAL("Hallo", netbuff.pop_data());
    BOOST_CHECK_EQUAL("Ola", netbuff.pop_data());
}

BOOST_AUTO_TEST_CASE(TestAppendingHalfWithOtherNetstring)
{
    NetstringBuffer netbuff = NetstringBuffer();
    netbuff.append_to_buf("5:Ha");
    netbuff.append_to_buf("llo,3:Ola,");
    BOOST_CHECK_EQUAL("Hallo", netbuff.pop_data());
    BOOST_CHECK_EQUAL("Ola", netbuff.pop_data());
}

BOOST_AUTO_TEST_CASE(TestAppendingHalfWithInvalidNetstring)
{
    NetstringBuffer netbuff = NetstringBuffer();
    netbuff.append_to_buf("5:Ha");

    BOOST_CHECK_THROW(netbuff.append_to_buf("llo3:Ola,"), netstring_error);
}

BOOST_AUTO_TEST_CASE(TestNestedNetstring)
{
    NetstringBuffer netbuff = NetstringBuffer();
    netbuff.append_to_buf("6:3:Ola,,");

    BOOST_CHECK_EQUAL("3:Ola,", netbuff.pop_data());

}

BOOST_AUTO_TEST_CASE(TestAllSplits)
{
    NetstringBuffer netbuff = NetstringBuffer();
    netbuff.append_to_buf("10:abcdefghij");
    netbuff.append_to_buf(",10:abcdefghi");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("j,10:abcdefgh");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("ij,10:abcdefg");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("hij,10:abcdef");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("ghij,10:abcde");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("fghij,10:abcd");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("efghij,10:abc");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("defghij,10:ab");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("cdefghij,10:a");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("bcdefghij,10:");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("abcdefghij,10");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf(":abcdefghij,1");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
    netbuff.append_to_buf("0:abcdefghij,");
    BOOST_CHECK_EQUAL("abcdefghij", netbuff.pop_data());
}

BOOST_AUTO_TEST_SUITE_END()
